list(
  # Adjustment name ----
  Name = "Reporting Delays",

  # Input parameters to the adjustment function ----
  Parameters = list(
    # Parameter 1: a specification with label, type and default value
    startYear = list(
      label = "Start year",
      value = 2000L,
      input = "numeric"),
    # Parameter 2: a specification with label, type and default value
    endQrt = list(
      label = "End quarter",
      value = 2017.25,
      input = "numeric"),
    stratVarNames = list(
      name = "stratVarNames",
      label = "Stratification columns (comma separated)",
      value = "",
      input = "text")
  ),

  # Names of packages that must be made available to the adjustment function ----
  RequiredPackageNames = c(),

  ## Adjustment function ----
  AdjustmentFunction = function(inputData, parameters) {

    # Color palette for the plots
    # First color: background of confidence interval for the estimated count
    # Second color: estimated count
    # Third color: reported count
    colorPalette <- c("#AAAAAA", "#E69F00", "#56B4E9")

    # Separator used for creating a composite of stratum columns. Should not occur in the stratum values.
    stratSep <- "_"

    # Start year
    startYear <- parameters$startYear
    # End quarter
    endQrt <- parameters$endQrt
    # Stratifiation columns
    stratVarNames <- trimws(unlist(strsplit(parameters$stratVarNames, ",")))
    stratVarNames <- stratVarNames[stratVarNames %in% colnames(inputData)]

    outputData <- copy(inputData)

    # B) PROCESS DATA -----------------------------------------------------------------------------

    # Create intermediate variables
    outputData[, ":="(
      NotificationTime = DateOfNotificationYear + 1/4 * DateOfNotificationQuarter,
      DiagnosisTime = DateOfDiagnosisYear + 1/4 * DateOfDiagnosisQuarter
    )]

    # Filter
    outputData <- outputData[DiagnosisTime > max(startYear, min(NotificationTime))
                             & NotificationTime <= endQrt]

    outputData[, VarX := 4 * (NotificationTime - DiagnosisTime)]
    outputData[VarX < 0, VarX := NA]
    outputData[, ":="(
      MinNotificationTime = min(NotificationTime),
      MaxNotificationTime = max(NotificationTime)
    ), by = .(ReportingCountry)]

    outputData[, ":="(
      VarT = 4 * (pmin(MaxNotificationTime, endQrt) - DiagnosisTime) + 1,
      Tf = 4 * (pmin(MaxNotificationTime, endQrt) - pmax(MinNotificationTime, startYear)),
      ReportingDelay = 1L
    )]
    outputData[, ":="(
      VarXs = Tf - VarX,
      VarTs = Tf - VarT
    )]
    # NOTE: Otherwise survival model complains
    outputData <- outputData[VarXs > VarTs]

    # Create a stratum variable - will be used to merge with the estimations,
    # takes missing categories as separate categories
    if (length(stratVarNames) > 0) {
      outputData[, Stratum := survival::strata(.SD, shortlabel = TRUE, sep = stratSep, na.group = TRUE),
                .SDcols = stratVarNames]
    } else {
      outputData[, Stratum := factor("Total")]
    }

    # Fit a stratified survival model
    model <- outputData[, survival::Surv(time = VarTs,
                                         time2 = VarXs,
                                         event = ReportingDelay)]
    fit <- outputData[, survival::survfit(model ~ Stratum)]
    if (is.null(fit$strata)) {
      strata <- 1L
    } else {
      strata <- fit$strata
    }

    # Aggregate and keep only required dimensions
    agregat <- outputData[, .(Count = .N),
                          by = .(DiagnosisTime,
                                 YearOfDiagnosis = DateOfDiagnosisYear,
                                 ReportingCountry,
                                 VarT,
                                 Stratum)]
    # Recreating stratum variables to assign them to the delay distribution dataset
    fitStratum <- data.table(
      Delay = fit$time,
      P = fit$surv,
      Var = fit$std.err^2,
      Stratum = factor(rep(seq_along(strata), strata),
                       labels = levels(agregat$Stratum)))
    if (length(stratVarNames) > 0) {
      fitStratum[, (stratVarNames) := tstrsplit(Stratum, stratSep)]
    }
    fitStratum[, VarT := max(Delay) - Delay]
    fitStratum <- fitStratum[VarT >= 0]
    # Merge fit data
    agregat <- fitStratum[agregat,
                          on = .(VarT, Stratum)]

    # Compute estimated count and its variance
    agregat[, ":="(
      EstCount = Count / P,
      EstCountVar = (Count * (Count + 1) / P^4 * Var) + Count * (1 - P) / P^2
    )]

    # c) TOTAL PLOT -------------------------------------------------------------------------------

    # Aggregating by year
    agregatByYear <- agregat[, .(Count = sum(Count),
                                 EstCount = sum(EstCount),
                                 EstCountVar = sum(EstCountVar)),
                             by = .(YearOfDiagnosis)]
    agregatByYear[, ":="(
      LowerEstCount = EstCount - 1.96 * sqrt(EstCountVar),
      UpperEstCount = EstCount + 1.96 * sqrt(EstCountVar)
    )]

    agregatPlotYr <- ggplot(agregatByYear, aes(x = YearOfDiagnosis)) +
      geom_ribbon(aes(ymin = pmin(pmax(LowerEstCount, 0), max(EstCount)),
                      ymax = pmin(pmax(UpperEstCount, 0), max(EstCount)),
                      fill = "95% confidence interval\nfor estimated total count"),
                  alpha = 0.4) +
      scale_fill_manual("Bounds", values = colorPalette[1]) +
      geom_line(aes(y = Count, color = "Reported"), size = 0.2) +
      geom_line(aes(y = EstCount, color = "Estimated total"), size = 0.5) +
      scale_colour_manual("Counts", values = colorPalette[2:3]) +
      ggtitle("Reported and estimated total count of cases") +
      xlab("Diagnosis year") +
      ylab("Count of HIV cases")

    # D) STRATIFIED PLOT (OPTIONAL) ---------------------------------------------------------------

    plotList <- NULL
    if (length(stratVarNames) > 0) {
      # Stratification
      colNames <- union(c("YearOfDiagnosis", "Count", "EstCount", "EstCountVar"),
                        stratVarNames)
      # Keep only required columns, convert data to "long" format...
      agregatByYearStrat <- melt(agregat[, ..colNames],
                                 measure.vars = stratVarNames,
                                 variable.name = "Stratum",
                                 value.name = "StratumValue")
      # ...and aggregate
      agregatByYearStrat <- agregatByYearStrat[, .(Count = sum(Count),
                                                   EstCount = sum(EstCount),
                                                   EstCountVar = sum(EstCountVar)),
                                               by = .(YearOfDiagnosis, Stratum, StratumValue)]
      agregatByYearStrat[, ":="(
        LowerEstCount = EstCount - 1.96 * sqrt(EstCountVar),
        UpperEstCount = EstCount + 1.96 * sqrt(EstCountVar),
        EstCountVar = NULL
      )]

      plotList <- lapply(stratVarNames, function(stratum) {
        ggplot(agregatByYearStrat[Stratum == stratum], aes(x = YearOfDiagnosis)) +
          geom_ribbon(aes(ymin = pmin(pmax(LowerEstCount, 0), max(EstCount)),
                          ymax = pmin(pmax(UpperEstCount, 0), max(EstCount)),
                          fill = "95% confidence interval\nfor estimated total count"),
                      alpha = 0.4) +
          scale_fill_manual("Bounds", values = colorPalette[1]) +
          geom_line(aes(y = Count, color = "Reported"), size = 0.2) +
          geom_line(aes(y = EstCount, color = "Estimated total"), size = 0.5) +
          scale_colour_manual("Counts", values = colorPalette[2:3]) +
          facet_wrap(~ StratumValue, ncol = 2) +
          ggtitle(paste("Reported and estimated total count of cases, by", stratum)) +
          xlab("Diagnosis year") +
          ylab("Count of HIV cases")
      })
    }

    artifacts <- list(OutputPlotYr = agregatPlotYr,
                      OutputPlotStrat = plotList)

    return(list(Table = agregat,
                Artifacts = artifacts))
  }
)
